require 'timeout'
require 'securerandom'
require 'uu/os/application/abstract_controller'
require 'uu/os/cmd/command_executor'
require 'uu/os/security/session'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'
require 'uu/os/util/container_service'
require 'uu/os/util/application_container_service'

module UU
  module OS
    module Util

      # Service providing system information of running module.
      # @private
      class CmdContainerService < UU::OS::Util::ApplicationContainerService # :nodoc:

        # @private
        def self.on_server_startup(ctx)
          UU::OS::Util::ContainerService.register_service(self.new, :on_health, :on_health_async, :on_health_callback)
        end

        # Returns health info.
        # @private
        def on_health(ctx) # :nodoc:
          result = super(ctx)
          result[:queueUri] = nil
          result[:taskUri] = nil
          entity = ctx.uu_uri
          start = Thread.current[:start]
          invocation = result[:trace].last
          if invocation.nil?
            invocation = {}
            result[:trace] << invocation
          end
          queue = ctx.parameters[:queueUri]
          if !queue.nil? && !queue.empty?
            result[:queueUri] = queue
            # Prepare asynchronous invocation.
            acmd_uc_code = URI.split(ctx.request.url)[5]
            acmd_uc_code.gsub!(/(^\/)|(\/$)/, '')
            parts = acmd_uc_code.split('/')
            action = parts.delete_at(-1)
            app_url_code = parts.delete_at(0)
            cmd_url_path = parts.join('/')
            params = {}
            params[:queue_uri] = queue
            params[:app_url_code] = app_url_code
            params[:cmd_url_path] = cmd_url_path
            params[:action] = action
            params[:parameters] = {result: result}
            # Prepare callback
            callback = {}
            callback[:appUrlCode] = app_url_code
            callback[:cmdUrlPath] = cmd_url_path
            callback[:action] = :healthCallback
            callback[:mainObjectUri] = entity
            params[:callback] = callback
            # Invoke asynchronous command
            session = UU::OS::Security::Session.current_session
            client = UU::OS::CMD::CommandExecutor.new(session)
            begin
              task_uri = client.invoke(entity, params)
              result[:taskUri] = task_uri
            rescue Exception => e
              invocation[:exception] = e.to_json
            end
            if !task_uri.nil?
              # Wait for finish to append health result.
              timeout = ctx.parameters[:timeout] || 120
              health = get_health(client, task_uri, timeout)
              if health[0].kind_of?(String)
                invocation[:exception] = health[0]
              else
                result[:trace].push(*health[0])
              end
              if !health[1].nil?
                result[:trace].push(*health[1])
              end
            end
          end
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return result
        end

        # Asynchronous part of health.
        # @private
        def on_health_async(ctx) # :nodoc:
          start = Thread.current[:start]
          invocation = {cont: 'ASYNC_CMD'}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return invocation
        end

        # Callback health.
        # @private
        def on_health_callback(ctx) # :nodoc:
          start = Thread.current[:start]
          invocation = {cont: 'CALLBACK'}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return invocation
        end

        # Returns hearbeat result of asynchronous invocation
        # (including health of callback).
        # @private
        def get_health(executor, task_uri, timeout) # :nodoc:
          begin
            Timeout::timeout(timeout.to_f) do
              state = executor.get_state(task_uri)
              while (state[:state] != 'FINISHED' && state[:state] != 'ERROR') ||
                  (state[:callbackState] != 'FINISHED' && state[:callbackState] != 'ERROR')
                sleep(1)
                state = executor.get_state(task_uri)
              end
              result = []
              if state[:state] == 'ERROR'
                result << state[:exception]
                return result
              else
                result << state[:result]
              end
              if state[:callbackState] == 'FINISHED'
                result << state[:callbackResult]
              end
              return result
            end
          rescue Timeout::Error
            result = []
            # Try to get actual state, maybe everything just finished
            state = executor.get_state(task_uri)
            if state[:state] != 'ERROR' && state[:state] != 'FINISHED'
              # Asynchronous command was not yet invoked
              result << UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::SERVER, nil, 'Asynchronous command was not yet started.').to_json
              return result
            elsif state[:state] == 'ERROR'
              result << state[:exception]
              return result
            else
              result << state[:result]
            end
            if state[:callbackState] != 'ERROR' && state[:callbackState] != 'FINISHED'
              # Callback was not yet invoked, append exception to async container
              result[0][-2][:exception] = UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::SERVER, nil, 'Callback was not yet started.').to_json
            elsif state[:callbackState] == 'FINISHED'
              result << state[:callbackResult]
            end
            return result
          end
        end


        def get_container_code(ctx)
          'CMD'
        end

      end

    end
  end
end
